/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.bm.event.mongo;

import java.util.List;

import org.alfresco.bm.event.AbstractEventService;
import org.alfresco.bm.event.Event;
import org.alfresco.bm.event.EventService;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.data.domain.Sort;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;

import com.mongodb.BasicDBObjectBuilder;
import com.mongodb.DBObject;

/**
 * An {@link EventService} <b>MongoDB</b> collection
 * 
 * @author Derek Hulley
 * @since 1.0
 */
public class MongoEventService extends AbstractEventService
{
    private static Log logger = LogFactory.getLog(MongoEventService.class);

    private final MongoTemplate mongo;
    private final String collection;

    /**
     * Construct a provider that works using the given Mongo instance and
     * collection name.
     * 
     * @param mongo the MongoDB template to use
     * @param collection the name of the MongoDB collection
     */
    public MongoEventService(MongoTemplate mongo, String collection)
    {
        this.mongo = mongo;
        this.collection = collection;
        // Initialize indexes
        DBObject idxNextAvailableEvent = BasicDBObjectBuilder
                .start(Event.FIELD_SCHEDULED_TIME, -1)
                .add(Event.FIELD_LOCK_OWNER, 1)
                .add(Event.FIELD_DATA_OWNER, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(idxNextAvailableEvent, Event.INDEX_NEXT_AVAILABLE_EVENT, false);
        
        DBObject idxLockOwner = BasicDBObjectBuilder
                .start(Event.FIELD_LOCK_OWNER, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(idxLockOwner, Event.INDEX_LOCK_OWNER, false);
        
        DBObject idxName = BasicDBObjectBuilder
                .start(Event.FIELD_NAME, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(idxName, Event.INDEX_NAME, false);
        
        DBObject uidxLockId = BasicDBObjectBuilder
                .start(Event.FIELD_LOCK_ID, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(uidxLockId, Event.INDEX_LOCK_ID, true);
    }

    @Override
    public long count()
    {
        return mongo.count(null, collection);
    }

    @Override
    public List<Event> findEvents(String name, int skip, int limit)
    {
        Criteria criteria = Criteria
                .where(Event.FIELD_NAME).is(name);
        Query query = new Query(criteria).skip(skip).limit(limit);
        
        // Query
        List<Event> events = mongo.find(query, Event.class, collection);
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Found events: \n" +
                    "   Event name: " + name + "\n" +
                    "   Events:     " + events);
        }
        return events;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Event nextEvent(String serverId, long latestScheduledTime, boolean localDataOnly)
    {
        if (serverId == null)
        {
            throw new IllegalArgumentException("'serverId' must be supplied.");
        }
        
        Criteria criteria = Criteria
                .where(Event.FIELD_LOCK_OWNER).is(null)
                .and(Event.FIELD_SCHEDULED_TIME).lte(new Long(latestScheduledTime));
        if (localDataOnly)
        {
            criteria = criteria.and(Event.FIELD_DATA_OWNER).is(serverId);
        }
        else
        {
            criteria = criteria.orOperator(
                    Criteria.where(Event.FIELD_DATA_OWNER).is(serverId),
                    Criteria.where(Event.FIELD_DATA_OWNER).is(null));
        }
        Query query = new Query(criteria);
        query.with(new Sort(Event.FIELD_SCHEDULED_TIME));

        long now = System.currentTimeMillis();
        Update update = new Update()
                .set(Event.FIELD_LOCK_OWNER, serverId)
                .set(Event.FIELD_LOCK_TIME, now);

        Event event = mongo.findAndModify(query, update, Event.class, collection);
        // Make sure we return the event, as modified
        if (event != null)
        {
            event.setLockOwner(serverId);
            event.setLockTime(now);
        }
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Fetched next event (no lock present): \n" +
                    "   Latest scheduled time:  " + latestScheduledTime + "\n" +
                    "   Server ID:              " + serverId + "\n" +
                    "   Event: " + event);
        }
        return event;
    }
    
    @Override
    public Event getEvent(String id)
    {
        Criteria criteria = Criteria.where(Event.FIELD_ID).is(id);
        Query query = new Query(criteria);
        return mongo.findOne(query, Event.class, collection);
    }

    @Override
    public boolean deleteEvent(Event event)
    {
        Criteria criteria = Criteria.where(Event.FIELD_ID).is(event.getId());
        Query query = new Query(criteria);
        Event deletedEvent = mongo.findAndRemove(query, Event.class, collection);
        if (deletedEvent == null)
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    @Override
    public Event putEvent(Event event)
    {
        mongo.insert(event, collection);
        // The event will be updated by the MongoTemplate
        return event;
    }
}
