/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.bm.session;

import java.util.Iterator;
import java.util.List;

import org.springframework.beans.factory.InitializingBean;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;

import com.mongodb.BasicDBObjectBuilder;
import com.mongodb.DBCollection;
import com.mongodb.DBObject;
import com.mongodb.WriteConcern;

/**
 * Mongo implementation of service providing access and management of {@link SessionData}.
 *
 * @author Derek Hulley
 * @since 1.4
 */
public class MongoSessionService extends AbstractSessionService implements InitializingBean
{
    private MongoTemplate mongo;
    private String collection;
    
    public MongoSessionService(MongoTemplate mongo, String collection)
    {
        this.mongo = mongo;
        this.collection = collection;
    }
    
    @Override
    public void afterPropertiesSet() throws Exception
    {
        checkIndexes();
    }

    /**
     * Ensure that the MongoDB collection has the required indexes
     */
    private void checkIndexes()
    {
        mongo.getDb().getCollection(collection).setWriteConcern(WriteConcern.SAFE);
        
        // Ensure unique session ID
        DBObject uidxSessionId = BasicDBObjectBuilder
                .start(SessionData.FIELD_SESSION_ID, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(uidxSessionId, "uidx_sessionId", true);

        // Ensure ordering by start time
        DBObject idxSessionId = BasicDBObjectBuilder
                .start(SessionData.FIELD_SESSION_ID, 1)
                .add(SessionData.FIELD_START_TIME, 2)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(idxSessionId, "idx_sessionId", false);

        // Find unfinished sessions
        DBObject idxEndTime = BasicDBObjectBuilder
                .start(SessionData.FIELD_END_TIME, 1)
                .add(SessionData.FIELD_START_TIME, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(idxEndTime, "idx_endTime", false);

        // Perhaps find long-running sessions
        DBObject idxElapsedTime = BasicDBObjectBuilder
                .start(SessionData.FIELD_ELAPSED_TIME, 1)
                .add(SessionData.FIELD_START_TIME, 1)
                .get();
        mongo.getDb().getCollection(collection).ensureIndex(idxElapsedTime, "idx_elapsedTime", false);
    }

    @Override
    protected void saveSessionData(SessionData sessionData)
    {
        mongo.insert(sessionData, collection);
    }

    @Override
    protected SessionData findSessionData(String sessionId)
    {
        Criteria criteria = Criteria.where(SessionData.FIELD_SESSION_ID).is(sessionId);
        Query query = new Query(criteria);

        SessionData sessionData = (SessionData) mongo.findOne(query, SessionData.class, collection);
        return sessionData;
    }

    @Override
    protected void updateSessionEndTime(String sessionId, long endTime, long elapsedTime)
    {
        Criteria criteria = Criteria.where(SessionData.FIELD_SESSION_ID).is(sessionId);
        Query query = new Query(criteria);

        Update update = Update
                .update(SessionData.FIELD_END_TIME, endTime)
                .set(SessionData.FIELD_ELAPSED_TIME, elapsedTime);
        mongo.updateFirst(query, update, collection);
    }

    @Override
    protected void updateSessionData(String sessionId, String data)
    {
        Criteria criteria = Criteria.where(SessionData.FIELD_SESSION_ID).is(sessionId);
        Query query = new Query(criteria);

        Update update = Update
                .update(SessionData.FIELD_DATA, data);
        mongo.updateFirst(query, update, collection);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Iterator<String> sessionIdIterator()
    {
        DBCollection sessionCollection = mongo.getCollection(collection);
        List<String> distinctSessionIds = sessionCollection.distinct(SessionData.FIELD_SESSION_ID);
        return distinctSessionIds.iterator();
    }

    @Override
    public long activeSessionsCount() {
        Criteria criteria = Criteria.where(SessionData.FIELD_END_TIME).is(-1);
        Query query = new Query(criteria);
        long count = mongo.count(query, collection);
        return count;
    }
}
